# ! /usr/bin/env python

import random
import socket
import time
from collections import defaultdict
from tempfile import mkdtemp

from contracts import contract
from contracts.utils import indent
from irc.bot import SingleServerIRCBot
from irc.client import ServerNotConnectedError, MessageTooLong

from . import __version__
from .dcache import create_propose_message, Envelope, CouldNotReadEnvelope
from .dcache_wire import create_ping_message
from .fragments import fragmentize, Assembler
from .ipfs_utils import IPFSInterface


class SwarmBot(SingleServerIRCBot):

    def __init__(self, servers, stream_name, mi, channel, nickname):
        SingleServerIRCBot.__init__(self, servers, nickname, nickname)
        self.mi = mi
        self.target = channel

        self.stream_name = stream_name
        self.seen = {}
        self.tmpdir = mkdtemp(prefix='swarm')

        self.admitted = False
        self.disconnected = False
        self.last_broadcast = 0

        self.lateness = random.uniform(0, 10)
        self.nusers = 10

        self.nreceived = 0
        self.assemblers = defaultdict(Assembler)

        ipfsi = mi.get_ipfsi()
        self.identity = ipfsi.ipfs_id()

    def on_pubmsg(self, c, e):
        self._handle(c, e)

    def on_privmsg(self, c, e):
        self._handle(c, e)

    def on_namreply(self, _c, e):
        names = map(str, e.arguments[2].split())
        print('list of handles: %s' % names)
        for name in names:
            e = Envelope(
                    maddr_from='/dswarm/%s/brain' % self.identity,
                    maddr_to='/dswarm/*/brain',
                    contents=create_ping_message())
            send(self, e, name)

    def _handle(self, _c, e):
        self.nreceived += 1
        s = str(e.arguments[0])
        # {'source': u'andrea!~andrea@nutonomy02.n.subnet.rcn.com',
        # 'tags': [], 'type': 'pubmsg', 'target': u'#duckiebots', 'arguments': [u'deplep']}
        source = e.source
        if '!' in source:
            source = source[:source.index('!')]

        sn = '/' + str(source)

        self.assemblers[sn].push(s)
        msgs = self.assemblers[sn].pop()
        for msg in msgs:
            try:
                envelope = Envelope.from_json(msg)
            except CouldNotReadEnvelope as e:
                em = 'Could not read envelope from %s: %s' % (sn, e)
                print(em)
            else:
                envelope.maddr_from = sn + envelope.maddr_from
                from duckietown_swarm.irc2 import CouldNotDispatch

                try:
                    self.mi.dispatch(envelope)
                except CouldNotDispatch as e:
                    print(s)

    def on_welcome(self, connection, _event):
        #        print('welcome message from %s: \n%s' % (connection.server, " ".join(event.arguments)))
        connection.join(self.target)
        connection.set_keepalive(30)
        print('Connected to %s on IRC server %s.' % (self.target, connection.server))
        self.admitted = True

    def on_join(self, c, e):
        nick = e.source.nick
        #        print('on_join %s' % nick)
        if nick != c.get_nickname():
            e = Envelope(
                    maddr_from='/dswarm/%s/brain' % self.identity,
                    maddr_to='/dswarm/*/brain',
                    contents=create_ping_message())
            send(self, e, nick)

    #            self.channels[ch] = Channel()
    #
    #        users = self.channels[self.target].users()
    #        self.nusers = len(users)
    #
    #        if False:
    #            msg = create_request_message(BOOTSTRAP_PATTERNS, validity=now_until(105))
    #            maddr_from = '/dswarm/%s/brain' % self.identity
    #            envelope = Envelope(maddr_from=maddr_from,
    #                         maddr_to='/dswarm/*/brain',
    #                         contents=msg)
    #            send(self, envelope, self.target)
    ##        print('current users: %s' % users)

    def on_disconnect(self, _c, event):
        print('Disconnected', event.__dict__)
        self.disconnected = True

    @contract(d=str)
    def send_message(self, target, d):
        self.connection.privmsg(target, d)
        #        print('to %s: %s' % (target, d))
        self.last_broadcast = time.time()


def start_irc(mi, server, channel):
    random.seed()

    ipfsi = IPFSInterface(mi.ipfs_path)
    identity = ipfsi.ipfs_id()
    hostname = socket.gethostname()
    printable = ''.join(ch for ch in hostname if ch.isalnum())
    printable = printable[:8]
    nickname = '%s_' % (__version__[-3:].replace('.', '')) + printable + '_' + identity[-4:]
    print('Using nickname %s' % nickname)

    delta_wait_for_welcome = 60
    stream_name = '/dns4/' + server[0] + '/tcp/' + str(server[1]) + '/irc'
    assert stream_name in mi.name2queue
    assert stream_name == mi.my_maddr

    #    key0 = CryptoKey(str(random.randint(0, 10000)), str(random.randint(0, 10000)))
    #    key0.hash = ipfsi.ipfs_id()

    t = time.time()
    irc_names = [
        stream_name + '/' + nickname,
        stream_name + '/' + channel,
    ]
    for irc_name in irc_names:
        msg = create_propose_message(('pri', identity, 'addresses'), irc_name, validity=[t, t + 300])
        mi.send_to_brain('', msg, sign=True)

    c = SwarmBot([server], stream_name, mi, channel, nickname)
    while True:

        while True:
            attempt_start = time.time()
            c.admitted = False
            print('server list: %s' % [_.host for _ in c.server_list])
            c._connect()
            while not c.admitted:
                # print('Waiting for welcome')
                c.reactor.process_once(1.0)

                delta = time.time() - attempt_start
                give_up = delta > delta_wait_for_welcome
                if give_up:
                    print('Could not receive welcome. giving up after %d s'
                          % delta_wait_for_welcome)
                    break
            if c.admitted:
                break
            print('Changing server')
            c.jump_server()

        c.disconnected = False

        while True:
            #            print('processing once')
            c.reactor.process_once()
            if c.disconnected:
                break

            xs = mi.get_many_for_me(timeout=1.0)
            #            print('get_many_for_me %s' % len(xs))
            for envelope in xs:
                try:
                    if not envelope.maddr_to or envelope.maddr_to[0] != '/':
                        msg = 'Invalid addr_to: %r' % envelope.maddr_to
                        msg += '\n\n' + indent(envelope, '  ')
                        print(msg)
                        continue

                    # orig = envelope.maddr_to
                    tokens = envelope.maddr_to.split('/')
                    assert tokens[0] == ''  # starts with '/'
                    dest = tokens[1]  # = '#duckiebots'
                    rest = '/' + "/".join(tokens[2:])
                    envelope.maddr_to = rest

                    #                    print('orig: %s  new: %s' % (orig, envelope.maddr_to))
                    if False:
                        print('not sending: %s' % envelope.contents)
                    else:
                        send(c, envelope, dest)

                #                    mi.send_to_brain(envelope.maddr_to, envelope.contents)
                except ServerNotConnectedError as e:
                    print e
                    break


@contract(envelope=Envelope)
def send(c, envelope, dest):
    envelope_json = envelope.to_json()
    #    print('send: %s' % envelope_json)
    limit, chunk_size = 300, 150
    try:
        if len(envelope_json) < limit:
            raw = [envelope_json]
        else:
            raw = fragmentize(envelope_json, max_chunk_length=chunk_size)

        for r in raw:
            if len(r) > 415:
                print('Do not expect a string of len %s to be ok.' % len(r))
            c.send_message(dest, r)
    except MessageTooLong as e:
        print(e)
        msg = 'Envelope size is %s' % len(envelope_json)
        print(msg)



import sys
import time
import traceback
from io import BytesIO
from multiprocessing import Process, Queue

from contracts import contract
from setproctitle import setproctitle  # @UnresolvedImport @UnusedImport
from termcolor import colored

from .utils import LineSplitter


class ProcessManager(object):
    status_queue = Queue()

    @contract(args='tuple')
    def __init__(self, f, args, name, restart):
        self.f = f
        self.args = args
        self.name = name
        self.restart = restart

    def start(self):
        args = (ProcessManager.status_queue, self.f, self.args, self.name, self.restart)
        self.p = Process(target=manager, args=args)
        self.p.daemon = True
        self.p.start()

    def join(self):
        self.process.join()


def manager(status_queue, f, args, name, restart):
    setproctitle('dt-swarm ' + name)

    nrestarts = 0
    min_restart_interval = 10
    restart_multiplier = 2.0

    def get_prefix():
        if nrestarts == 0:
            return colored(name, attrs=['dark'])
        else:
            return colored(name + '!%d' % nrestarts, 'red')

    t1 = lambda s: '%s|%s' % (get_prefix(), s)
    original_stdout = sys.stdout
    sys.stdout = AddPrefixStream(original_stdout, t1)

    status_queue.put(('starting', name))
    while True:
        last_restart = time.time()
        try:
            f(*args)
        except Exception as e:
            nrestarts += 1
            min_restart_interval *= restart_multiplier
            print(traceback.format_exc(e))
            if not restart:
                break
        else:
            print('graceful exit')
            break
        now = time.time()
        wait_until = last_restart + min_restart_interval
        if now < wait_until:
            wait = wait_until - now
            print('waiting %d s before starting' % wait)
            time.sleep(wait)

        status_queue.put(('restart', name))

    status_queue.put(('exit', name))
    print('peaceful exit')


class AddPrefixStream(object):

    def __init__(self, to, transform):
        self.buffer = BytesIO()
        self.line_splitter = LineSplitter()
        self.to = to
        self.transform = transform

    def fileno(self):
        return 0

    def write(self, s):
        self.line_splitter.append_chars(s)
        lines = self.line_splitter.lines()
        for line in lines:
            self.to.write(self.transform(line) + '\n')
        self.to.flush()

    def flush(self):
        pass



from contracts import contract
from contracts.utils import indent
from system_cmd import CmdException

from .brain import DistributedRepoSwarm
from .dcache_wire import CouldNotInterpret, interpret_message
from .ipfs_utils import Timeout
from .irc2 import MessagingInterface

seen_mh = set()
seen = set()


@contract(mi=MessagingInterface)
def read_nodes_summaries(mi):
    timeout = '60s'
    ipfsi = mi.get_ipfsi()

    while True:
        #        print('waiting for summary')
        envelopes = [mi.get_next_for_me()]
        #
        n = len(envelopes)
        if n > 1:
            print('ignoring %s messages and process the last' % (n - 1))
            envelopes = [envelopes[-1]]

        for envelope in envelopes:
            #
            #            if envelope.maddr_from == mi.MADDR_BRAIN:
            #                print('skipping this from ourselves: %s' % envelope)
            #                continue

            #            print envelope

            mh = envelope.contents
            if mh in seen_mh:
                continue

            seen_mh.add(mh)
            data_mh = mh + '/machines.txt'
            print('getting %s' % data_mh)

            try:
                data = ipfsi.cat(data_mh, timeout=timeout)
            #                print('last chars of data: %r' % data[-300:])
            except Timeout:
                print('timeout for %s' % data_mh)
                continue
            except CmdException as e:
                print(str(e))
                continue

            data = data.strip()
            if not data:
                msg = 'Could not find any data in %s' % data_mh
                print(msg)
                continue

            lines = data.split('\n')

            i = 0
            for k, msg in enumerate(lines):
                if msg in seen:
                    continue

                #                    print('NEW: %s' % msg)

                try:
                    interpret_message(msg)
                except CouldNotInterpret as e:
                    m = 'Could not interpret line %d of %d in %s' % (k, len(lines), data_mh)
                    m += '\n\n' + indent(msg.__repr__(), ' line > ')
                    m += '\n\n' + indent(str(e), '> ')
                    print(m)
                    continue

                seen.add(msg)
                i += 1
                from_channel = '/ipfs/' + mh
                mi.send_to_brain(from_channel, msg)

            new_ones = i
            print('read %s catchup messages: %s new ones' % (len(lines), new_ones))


def read_lines(s):
    lines = s.split('\n')
    dc = DistributedRepoSwarm()
    from_channel = 'stdin'
    i = 0
    for msg in lines:
        dc.process(msg, from_channel)
        i += 1
    print('processed: %s' % i)
    return dc



'''

{'fragment': 0, 'n': 10, 'sha256_base58': 'ckodokocdk', 'chunk_base58': ''}


'''
import json
from collections import OrderedDict, defaultdict

from base58 import b58encode, b58decode
from contracts import contract
from contracts.utils import raise_desc

from .dcache import get_sha256_base58


@contract(s0=str, max_chunk_length='int', returns='list(str)')
def fragmentize(s0, max_chunk_length):
    # first, divide in chunks
    chunks = []
    s = s0
    while len(s) > 0:
        chunk = s[:max_chunk_length]
        chunks.append(chunk)
        s = s[max_chunk_length:]

    assert ''.join(chunks) == s0, (chunks, s0)

    checksum = get_sha256_base58(s0)
    msgs = []
    for i, chunk in enumerate(chunks):
        od = OrderedDict()
        od['fragment'] = i
        od['n'] = len(chunks)
        od['sha256_base58'] = checksum
        od['chunk_base58'] = b58encode(chunk)
        j = json.dumps(od)
        msgs.append(j)
    return msgs


class Assembler(object):

    def __init__(self):

        # sha -> i -> message
        self.chunks = defaultdict(lambda: {})
        self.output = []

    def push(self, s):
        try:
            j = json.loads(s)
        except ValueError:
            msg = 'Cannot read string of size %s: %r' % (len(s), s)
            print(msg)
            return

        if 'fragment' in j:
            self.chunks[j['sha256_base58']][j['fragment']] = j
        else:
            self.output.append(s)
        self.check_complete()

    def pop(self):
        res = self.output
        self.output = []
        return res

    def check_complete(self):
        for c, fragment2chunks in list(self.chunks.items()):
            try:
                s = assemble_chunks(fragment2chunks)
            except Corrupted as e:
                print(e)
                self.chunks.pop(c)
            except Incomplete:
                pass
            else:
                self.output.append(s)
                self.chunks.pop(c)


class Corrupted(Exception):
    pass


class Incomplete(Exception):
    pass


def assemble_chunks(fragment2chunk):
    one = list(fragment2chunk.values())[0]
    n = one['n']
    s = ''
    for i in range(n):
        if not i in fragment2chunk:
            raise Incomplete()

        chunk = fragment2chunk[i]
        si = b58decode(str(chunk['chunk_base58']))
        s += si

    checksum = get_sha256_base58(s)
    if checksum != one['sha256_base58']:
        raise_desc(Corrupted, 'Failed checksum', s=s.__repr__())
    return s



import asyncore
import socket
from abc import ABCMeta, abstractmethod

from .dcache import Envelope, CouldNotReadEnvelope
from .utils import MakeLines


# def definitely_send(socket, s):
#    while s:
#        try:
#            n = socket.send(s)
#            s = s[n:]
#            break
#        except IOError as e:
#            if e.errno == errno.EAGAIN:
#                sys.stderr.write('E')
#                time.sleep(0.1)
#                continue
#            if e.errno == 98:
#                print('broken pipe')
#                raise
def tcplisten(mi, port):
    server = EchoServer(mi, '0.0.0.0', port)
    while 1:
        asyncore.loop(count=1, timeout=1.0)
        for e in mi.get_all_available_for_me():
            for addr, socket in server.addr2socket.items():
                if e.maddr_to.startswith(addr):
                    rest = e.maddr_to[len(addr):]
                    e.maddr_to = rest
                    s = e.to_json() + '\n'
                    try:
                        socket.setblocking(1)
                        socket.sendall(s)
                    except IOError as e:
                        print('Could not send packet: ' + e)
                    break
            else:
                msg = 'Could not route connection %s' % e.maddr_to
                print(msg)


class Processor():
    __metaclass__ = ABCMeta

    def set_handler(self, handler):
        self.handler = handler

    @abstractmethod
    def receive(self, line):
        pass

    def send(self, line):
        self.handler.send(line)


class EchoHandler(asyncore.dispatcher_with_send):

    def init(self, processor):
        self.processor = processor
        self.ml = MakeLines()

    def handle_read(self):
        s = self.recv(8192)
        #        print('handle_read : %r' % s)
        self.ml.push(s)
        for line in self.ml.get_lines():
            #            print('line : %r' % line)
            self.processor.receive(line)


class EchoServer(asyncore.dispatcher):

    def __init__(self, mi, host, port):
        self.mi = mi
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind((host, port))
        self.listen(5)
        self.addr2socket = {}

    def handle_accept(self):
        pair = self.accept()

        if pair is not None:
            sock, addr = pair
            sock.setblocking(1)
            maddr = '/ip4/%s/tcp/%s' % addr
            self.addr2socket[maddr] = sock
            print('Incoming connection from %s' % maddr)
            handler = EchoHandler(sock)
            processor = MyProcessor(self.mi, addr)
            processor.set_handler(handler)
            handler.init(processor)


class MyProcessor(Processor):

    def __init__(self, mi, addr):
        self.mi = mi
        self.host, self.port = addr

    def receive(self, line):
        try:
            env = Envelope.from_json(line)
        except CouldNotReadEnvelope as e:
            print(str(e))
            self.send(str(e) + '\n')
        else:
            m = '/ip4/%s/tcp/%s' % (self.host, self.port)
            env.maddr_from = m + env.maddr_from
            print('dispatching %s' % env)
            self.mi.dispatch(env)



from .constants import DSwarmConstants
from .dcache_wire import create_dismiss_message, create_propose_message
from .ipfs_utils import CouldNotPin, IPFSInterface, InvalidHash
from .packaging import find_more_information, UnexpectedFormat
from .utils import as_seconds, now_until, now_until_forever


def verifier(mi):
    ipfsi = IPFSInterface(mi.ipfs_path)
    done = set()

    while True:
        mh = mi.get_next_for_me().contents
        #        print('considering %s' % mh)
        if mh in done:
            #            print('verifier received %s twice' % mh)
            continue

        done.add(mh)

        try:
            more_info = find_more_information(ipfsi, mh, find_provs=False)
            print('%s verified %s' % (mh, more_info.filename))
        except InvalidHash as _e:
            validity = now_until(as_seconds('1h'))
            m = create_dismiss_message(DSwarmConstants.DLC_BUCKET_FILES, mh, validity)
            mi.send_to_brain('', m)
            continue
        except UnexpectedFormat as _e:
            print('Could not find more information for %s' % mh)
            validity = now_until(as_seconds('1h'))
            m = create_dismiss_message(DSwarmConstants.DLC_BUCKET_FILES, mh, validity)
            mi.send_to_brain('', m)
            continue

        validity = now_until_forever()
        m = create_propose_message(DSwarmConstants.DLC_BUCKET_VERIFIED, mh, validity)
        mi.send_to_brain('', m, sign=True)


def pinner(mi):
    timeout = '15m'
    ## Read from the queue
    ipfsi = IPFSInterface(mi.ipfs_path)
    done = set()
    while True:
        mh = mi.get_next_for_me().contents

        if mh in done:
            print('Pinner received %s twice' % mh)
            continue

        done.add(mh)

        try:
            ipfsi.pin_add(mh, recursive=True, timeout=timeout)
        except CouldNotPin as e:
            print('Pinner %s: could not pin file: %s' % (mh, e))
        else:
            print('Pinner %s: OK' % (mh))

            validity = now_until(as_seconds('7d'))
            m = create_propose_message(DSwarmConstants.DLC_BUCKET_SAFE, mh, validity)
            mi.send_to_brain('', m, sign=True)



from setuptools import find_packages, setup


def get_version(filename):
    import ast
    version = None
    with open(filename) as f:
        for line in f:
            if line.startswith('__version__'):
                version = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError('No version found in %r.' % filename)
    if version is None:
        raise ValueError(filename)
    return version


version = get_version(filename='src/duckietown_swarm/__init__.py')

name = 'duckietown_swarm'

setup(name=name,
      url='http://github.com/duckietown/duckietown_swarm',
      maintainer="Andrea Censi",
      maintainer_email="andrea@duckietown.org",
      description='',
      long_description='',
      # package_data={'':['*.*', '*.mcdp*', '*.js', '*.png', '*.css']},

      # without this, the stuff is included but not installed
      include_package_data=True,
      keywords="Optimization",
      license="GPLv2",
      classifiers=[
          'Development Status :: 4 - Beta',
      ],
      version=version,

      download_url=
      'http://github.com/duckietown/duckietown_swarm/tarball/%s' % version,

      package_dir={'': 'src'},
      packages=find_packages('src'),
      install_requires=[
          # 'ConfTools>=1.0,<2',
          # '#quickapp',
          # 'reprep',
          'irc',
          'SystemCmd',
          'ConfTools',
          'termcolor',
          'ruamel.yaml',
          'ruamel.ordereddict',
          'base58',
          'netifaces',
          'multiaddr',
          'subprocess32',
          'setproctitle',
          'networkx>=1.10,<2',
      ],
      # This avoids creating the egg file, which is a zip file, which makes our data
      # inaccessible by dir_from_package_name()
      zip_safe=False,
      dependency_links=[
      ],

      tests_require=[
          #        'nose>=1.1.2,<2',
          #        'comptests',
      ],

      entry_points={

          'console_scripts': [
              'dt-swarm = duckietown_swarm:duckietown_swarm_cli_main',
              'dt-swarm-watcher = duckietown_swarm:duckietown_swarm_watcher_main',
          ]
      }
      )



from .constants import DSwarmConstants

from .dcache import Envelope, interpret_message
from .dcache_wire import create_ping_message
from .socket_utils import AsyncLineBasedConnection


def duckietown_swarm_ping_main():
    import argparse

    parser = argparse.ArgumentParser(description='Ping the swarm')
    parser.add_argument('host', type=str, default='localhost')
    parser.add_argument('who', default='/dswarm/*/brain', help='who to ping')

    args = parser.parse_args()

    host = args.host
    port = DSwarmConstants.port_dswarm
    maddr_to = args.who
    albc = AsyncLineBasedConnection(host, port)
    maddr_from = '/cli'

    inbound = albc.get_q_inbound()
    outbound = albc.get_q_outbound()

    def ping():
        e = Envelope(maddr_from=maddr_from,
                     maddr_to=maddr_to,
                     contents=create_ping_message())
        outbound.put(e)

    ping()

    while True:
        env = inbound.get()
        msg = interpret_message(env.contents)
        print('obtained from %s: %s' % (env.maddr_from, msg.__repr__()))
        ping()


if __name__ == '__main__':
    duckietown_swarm_ping_main()



import os

from system_cmd import system_cmd_result


def shell_access(port):
    script = '''#!/bin/bash
echo TERM=$TERM
echo Starting the shell > /dev/stderr
echo
echo Welcome to `hostname`
echo
echo If backspace does not work, you should call on the client the following command:
echo
echo     $  stty erase '^h'
echo
echo Note that ctrl-c does not work.
echo
echo
/bin/bash -i 2>&1
'''
    filename = '/tmp/shell'
    with open(filename, 'w') as f:
        f.write(script)

    os.chmod(filename, 0744)
    print('starting ncat')
    cmd = ['ncat', '--allow', '127.0.0.1', '-k', '-l', '-p', str(port), '-e', filename]
    system_cmd_result('.', cmd, raise_on_error=True)
#    p = subprocess.Popen(
#                cmd,
#                stdout=sys.stdout,
#                stderr=sys.stderr,
#                bufsize=0,
#
#                cwd='.')
#
#    p.wait()



# from datetime import datetime
# import socket
#
# from system_cmd import CmdException
#
# from .ipfs_utils import IPFSInterface, InvalidHash
# from .ipfs_utils import MakeIPFS2
# from .packaging import find_more_information, UnexpectedFormat
from system_cmd import CmdException

from .ipfs_utils import IPFSInterface


# def publisher_summary(mi, channel_name):
#    KEY_NAME = 'summary'
#    known = set()
#
#    ipfsi = IPFSInterface()
#    keys = ipfsi.get_keys()
#    path = '/ipns/' + keys[KEY_NAME]
#    print('You will see my summaries at %s' % path)
#
#    ipfsi = IPFSInterface()
#
#    previous_hash = None
#    while True:
#        timeout = 5
#
#        hashes = mi.get_at_least_one_publisher(timeout)
#        changes = 0
#        for x in hashes:
#            what, ipfs_hash = x
#            if what == 'add':
#                if not ipfs_hash in known:
#                    print('publisher: adding %s' % ipfs_hash)
#                    known.add(ipfs_hash)
#                    changes += 1
#            elif what == 'remove':
#                if ipfs_hash in known:
#                    known.remove(ipfs_hash)
#                    changes += 1
#                    print('publisher: remove %s' % ipfs_hash)
#            else:
#                assert False, x
#
#        if changes == 0:
#            continue
#        #print('get_at_least_one: %.1f s' % (time.time() - t0))
#
#        print('Creating summary for %d more, %d changes' % (len(known), changes))
#
#        summary_hash = get_summary(ipfsi, known, permanent=path, out_queues=write_to)
#        print('summary: %s' % summary_hash)
#
#        if previous_hash != summary_hash:
#            print('summary hash: %s' % summary_hash)
#
#            mi.send_to_ipns_publisher('summary', summary_hash)
#
#            previous_hash = summary_hash
#
#            msg = {'mtype': 'summary', 'details': {'ipfs': summary_hash}}

#            for q in write_to:
#                q.put((KEY_NAME, msg), block=False)


def publish_ipns(mi):
    ipfsi = IPFSInterface(mi.ipfs_path)
    keys = ipfsi.get_keys()

    while True:
        # print('summary')
        timeout = 5
        what = mi.get_many_for_me(timeout=timeout)
        if what:
            to_add = {}
            for envelope in what:
                key, v = envelope.contents
                to_add[key] = v

            for key, v in to_add.items():

                if not key in keys:
                    print('Generating new key %r' % key)
                    ipfsi.gen_key(key, 'rsa', 2048)
                    keys = ipfsi.get_keys()
                try:
                    print('publishing to key %r: %s' % (key, v))
                    ipfsi.publish(key, v)
                except CmdException as e:
                    print(e)  # XXX
                else:
                    path = '/ipns/' + keys[key]
                    print('ipns: published with key %s  %s -> /ipfs/%s ' % (key, path, v))

# def get_summary(ipfsi, known, permanent, out_queues):
#    from duckietown_swarm.pinner import advertise_invalid
#    m = MakeIPFS2()
#    s = '<html><head></head><body><pre>\n'
#
#    s += '\n Created: %s' % str(datetime.now())[:16]
#    s += '\n Host: %s' % socket.gethostname()
#    s += '\n Permanent path: <a href="%s">%s</a>' % (permanent, permanent)
#
#    s += '\n'
#    found = []
#    others = []
#    invalid = []
#    for k in known:
#        try:
#            f = find_more_information(ipfsi, k)
#            found.append(f)
#        except UnexpectedFormat:
#            others.append(k)
#        except InvalidHash:
#            invalid.append(k)
#
#    s += '<h2>Good uploads</h2>'
#    already = set()
#    redundant = []
#    found = sorted(found, key=lambda _: _.ctime)
#
#    def description(f):
#        ss = ''
#        ss += '<a href="/ipfs/%s">info</a>' % (f.ipfs_info)
#        ss += ' <a href="/ipfs/%s">raw</a>' % (f.ipfs)
#        ss += ' providers %3d %3d' % (len(f.providers_info), len(f.providers_payload))
#        ss += ' ' + str(f.ctime)[:16]
#        ss += ' %5d MB' % (f.size / (1000.0 * 1000))
#        ss += ' <a href="/ipfs/%s">%s</a>' % (f.ipfs_payload, '%20s' % f.filename)
#        ss += ' uploaded by %s : %s @ %s' % (f.upload_node, f.upload_user, f.upload_host)
#        return ss
#
#    for f in found:
#        if f.ipfs_payload in already:
#            redundant.append(f)
#            continue
#        already.add(f.ipfs_payload)
#        m.add_file(f.ipfs_payload, f.ipfs_payload, f.size)
#        m.add_file(f.ipfs_info, f.ipfs_info, 0)
#        s += '\n' + description(f)
#
#    if redundant:
#        s += '<h2>Different uploads of same payload</h2>'
#        for h in redundant:
#            s += '\n' + description(h)
#
#            advertise_invalid(out_queues, h.ipfs, comment="redundant")
#
#    if others:
#        s += '<h2>Other uploads</h2>'
#        for h in others:
#            m.add_file(h, h, 0)
#            s += '\n<a href="%s">%s</a>' % (h, h)
#
#    if invalid:
#        s += '<h2>Invalid hashes</h2>'
#        for h in invalid:
#            s += '\n' + h
#
#    s += '\n</pre></body>'
#
#    m.add_file_content('index.html', s)
#
#    return m.get_ipfs_hash()
